/*
 * main.cpp - match package name with cpe dictionary data
 *
 * Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>

#include <algorithm>
#include <cctype>
#include <fstream>
#include <regex>
#include <set>
#include <stdexcept>
#include <string>
#include <tuple>
#include <utility>

#include <boost/algorithm/string.hpp>

#include <boost/format.hpp>

#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>

#include <clickhouse-cpp/client.h>

#include "helpers.h"

static boost::format create_format(const std::string &f_string)
{
    boost::format fmter(f_string);
    fmter.exceptions(boost::io::all_error_bits ^ (boost::io::too_many_args_bit | boost::io::too_few_args_bit));
    return fmter;
}

static void show_version()
{
        const char *msg =
            "\
" PACKAGE " " PACKAGE_VERSION
            "\n\
Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>\n\
" PACKAGE_NAME
            " is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.";
        fprintf(stderr, "%s\n", msg);
}

int main(int argc, char **argv)
{
    try
    {
        bool show_help = false;
        bool _show_version = false;
        bool verbose = false;
        std::string clickhouse_config_file;
        std::string rules_filename;

        boost::program_options::options_description desc("General Options");
        desc.add_options()
            ("help,h", boost::program_options::bool_switch(&show_help)->default_value(false), "Show help")
            ("version,v", boost::program_options::bool_switch(&_show_version)->default_value(false), "Show version")
            ("verbose,V", boost::program_options::bool_switch(&verbose)->default_value(false), "Very verbose mode")
            ("clickhouse-config,Y", boost::program_options::value< std::string >(&clickhouse_config_file), "Path to file containing configuration for connection to clickhouse database");

        boost::program_options::options_description hidden;
        hidden.add_options()
            ("rules-filename", boost::program_options::value< std::string >(&rules_filename)->required(), "rules matching filename");

        boost::program_options::options_description all_options;
        all_options.add(desc);
        all_options.add(hidden);

        boost::program_options::positional_options_description positional_options;
        positional_options.add("rules-filename", 1);

        boost::program_options::variables_map vm;

        boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(positional_options).run(), vm);
        boost::program_options::notify(vm);

        if (show_help) {
                std::stringstream descr;
                descr << all_options;
                printf("Usage: %s [options]\n%s\n", argv[0], descr.str().c_str());
                return EXIT_SUCCESS;
        }

        if (_show_version) {
                show_version();
                return EXIT_SUCCESS;
        }

        if (clickhouse_config_file.empty()) {
                fprintf(stderr, "Clickhouse config is required\n");
                return EXIT_FAILURE;
        }

        if (rules_filename.empty()) {
            fprintf(stderr, "CPE matching rules file is required\n");
            return EXIT_FAILURE;
        }

        // product -> vendor -> software_platform
        std::map<std::string, std::map<std::string, std::set<std::string> > > products_map;

        if (verbose) {
            fprintf(stderr, "Obtaining information about CPE dictionary entries\n");
        }

        clickhouse::ClientOptions client_options, client_options_readonly;

        std::string table_name_cpe_dictionary = "CpeDictionary";
        std::string table_name_cpe_dictionary_matches = "CpeDictionaryMatches";
        std::string table_name_packages = "Packages";

        /* First read config */
        {
            boost::property_tree::ptree clickhouse_config_tree;

            std::ifstream stream(clickhouse_config_file);
            boost::property_tree::read_ini(stream, clickhouse_config_tree);

            auto table_names = clickhouse_config_tree.get_child_optional("Tables");
            if (table_names)
            {
                optional_set(table_name_cpe_dictionary, *table_names, "TableNameCpeDictionary", &is_valid_table_name);
                optional_set(table_name_cpe_dictionary_matches, *table_names, "TableNameCpeDictionaryMatches", &is_valid_table_name);
                optional_set(table_name_packages, *table_names, "TableNamePackage", &is_valid_table_name);
            }

            auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
            if (output_connection)
            {
                fill_connection_options(*output_connection, client_options);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options);
                }
            }

            auto input_connection = clickhouse_config_tree.get_child_optional("ConnectionReadOnly");
            if (input_connection)
            {
                fill_connection_options(*input_connection, client_options_readonly);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options_readonly);
                }
            }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        std::string query_string;

        /* Create tables if they're missing */
        client.Execute(cpe_dictionary_table_creation_command(table_name_cpe_dictionary));

        if (verbose) {
            fprintf(stderr, "Obtaining information about existing CPEs\n");
        }

        try
        {
            query_string = std::string(
                        "SELECT cd_vendor"
                        ", cd_product"
                        ", cd_software_platform"
                        " FROM ("
                            "SELECT cd_vendor"
                            ", cd_product"
                            ", cd_software_platform"
                            ", argMax(cd_status, ts) AS cd_status_ts"
                            " FROM ") + table_name_cpe_dictionary + " GROUP BY cd_vendor, cd_product, cd_software_platform"
                        ") WHERE cd_status_ts = 'actual'";

            /* Now fetch all source packages data */
            client.Select(query_string,
                          [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    auto cd_vendor = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                    auto cd_product = sv_to_string(ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i));
                    auto cd_software_platform = sv_to_string(ensure_not_null(block[2]->As<clickhouse::ColumnString>())->At(i));

                    products_map[cd_product][cd_vendor].insert(cd_software_platform);
                }
            });

            query_string.clear();
        }
        catch (...)
        {
            if (!query_string.empty())
            {
                fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
            }

            throw;
        }

        std::vector<std::tuple<std::string, std::string, std::string, std::string, std::set<std::string> > > rules;

        if (verbose) {
            fprintf(stderr, "Reading matching rules\n");
        }

        {
            std::ifstream rules_file_stream(rules_filename);
            std::string rule_line;

            while (std::getline(rules_file_stream, rule_line))
            {
                if (rule_line.empty() || (rule_line[0] == '#'))
                {
                    continue;
                }

                std::vector<std::string> rules_array;
                boost::split(rules_array, rule_line, boost::is_any_of(";"));

                if (rules_array.size() != 5)
                {
                    fprintf(stderr, "Invalid rule line: %s\n", rule_line.c_str());
                    continue;
                }

                std::set<std::string> flags;
                boost::split(flags, rules_array[4], boost::is_any_of("|"));

                rules.push_back(std::move(std::make_tuple(std::move(rules_array[0]), std::move(rules_array[1]), std::move(rules_array[2]), std::move(rules_array[3]), std::move(flags))));
            }
        }

        if (verbose) {
            fprintf(stderr, "Obtaining package names\n");
        }

        std::set<std::string> pkgnames;

        try
        {
            query_string = std::string("SELECT distinct(pkg_name) FROM " + table_name_packages);

            clickhouse::Client client_readonly(client_options_readonly);

            /* Now fetch all source packages data */
            client_readonly.Select(query_string,
                          [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    std::string pkgname = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                    pkgnames.insert(std::move(pkgname));
                }
            });

            query_string.clear();
        }
        catch (...)
        {
            if (!query_string.empty())
            {
                fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
            }

            throw;
        }

        client.Execute(cpe_dictionary_match_table_creation_command(table_name_cpe_dictionary_matches));

        // pkgname -> ( vendor, product, software platform )
        std::map<std::string, std::set<std::tuple<std::string, std::string, std::string> > > old_matches, new_matches;

        if (verbose)
        {
            fprintf(stderr, "Reading old matches matches\n");
        }

        try
        {
            query_string = std::string(
                        "SELECT pkg_name"
                        ", argMax(cdm_match.vendor, ts)"
                        ", argMax(cdm_match.product, ts)"
                        ", argMax(cdm_match.software_platform, ts)"
                        " FROM " + table_name_cpe_dictionary_matches + " GROUP BY pkg_name");

            /* Now fetch all source packages data */
            client.Select(query_string,
                          [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    auto pkg_name = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));

                    auto cdm_vendor = ensure_not_null(ensure_not_null(block[1]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto cdm_product = ensure_not_null(ensure_not_null(block[2]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto cdm_software_platform = ensure_not_null(ensure_not_null(block[3]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());

                    if ((cdm_vendor->Size() != cdm_product->Size())
                            || (cdm_vendor->Size() != cdm_software_platform->Size()))
                    {
                            std::stringstream str;
                            str << "Columns cdm_match.vendor, cdm_match.product, cdm_match.software_platform have mismatching sizes!";
                            throw std::runtime_error(str.str());
                    }

                    for (size_t j = 0; j < cdm_vendor->Size(); ++j)
                    {
                        old_matches[pkg_name].insert(std::make_tuple(sv_to_string(cdm_vendor->At(j)), sv_to_string(cdm_product->At(j)), sv_to_string(cdm_software_platform->At(j))));
                    }
                }
            });

            query_string.clear();
        }
        catch (...)
        {
            if (!query_string.empty())
            {
                fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
            }

            throw;
        }

        if (verbose)
        {
            fprintf(stderr, "Processing matches\n");
        }

        for (auto pkg_iter = pkgnames.begin(); pkg_iter != pkgnames.end(); ++pkg_iter)
        {
            std::string vendor;
            std::string product = *pkg_iter;
            std::string software_platform;

            // all CPE records are lowercase
            std::transform(product.begin(), product.end(), product.begin(), [](unsigned char c){ return std::tolower(c); });

            new_matches[*pkg_iter] = std::set<std::tuple<std::string, std::string, std::string> >();

            auto check_product_iter = products_map.find(*pkg_iter);
            if (check_product_iter != products_map.end())
            {
                if (verbose)
                {
                    fprintf(stderr, "Package \'%s\' matched to \'%s\':\'%s\'\n", pkg_iter->c_str(), "", pkg_iter->c_str());
                }

                new_matches[*pkg_iter].insert(std::make_tuple(vendor, product, software_platform));
            }

            for (auto rule_iter = rules.begin(); rule_iter != rules.end(); ++rule_iter)
            {
                std::regex reg(std::get<0>(*rule_iter));
                std::smatch match;

                if (std::regex_match(product, match, reg))
                {
                    auto vendor_fmt = create_format(std::get<1>(*rule_iter));
                    auto product_fmt = create_format(std::get<2>(*rule_iter));
                    auto software_platform_fmt = create_format(std::get<3>(*rule_iter));

                    vendor_fmt = vendor_fmt % vendor;
                    product_fmt = product_fmt % product;
                    software_platform_fmt = software_platform_fmt % software_platform;

                    for (auto i = 1; i < match.size(); ++i)
                    {
                        vendor_fmt = vendor_fmt % match[i];
                        product_fmt = product_fmt % match[i];
                        software_platform_fmt = software_platform_fmt % match[i];
                    }

                    std::string vendor_tmp = vendor_fmt.str();
                    std::string product_tmp = product_fmt.str();
                    std::string software_platform_tmp = software_platform_fmt.str();

                    auto flags = std::get<4>(*rule_iter);
                    if (flags.find("temp") == flags.end())
                    {
                        vendor = vendor_tmp;
                        product = product_tmp;
                        software_platform = software_platform_tmp;
                    }

                    bool matched = false;

                    auto product_iter = products_map.find(product_tmp);
                    if (product_iter != products_map.end())
                    {
                        for (auto vendor_iter = product_iter->second.begin(); vendor_iter != product_iter->second.end(); ++vendor_iter)
                        {
                            if ((!vendor_tmp.empty()) && (vendor_iter->first != vendor_tmp))
                            {
                                continue;
                            }

                            if (software_platform_tmp.empty() || (vendor_iter->second.find(software_platform_tmp) != vendor_iter->second.end()))
                            {
                                if (verbose)
                                {
                                    fprintf(stderr, "Package \'%s\' matched to \'%s\':\'%s\':\'%s\'\n", pkg_iter->c_str(), vendor_tmp.c_str(), product_tmp.c_str(), software_platform_tmp.c_str());
                                }

                                new_matches[*pkg_iter].insert(std::make_tuple(vendor_tmp, product_tmp, software_platform_tmp));

                                matched = true;
                            }
                        }
                    }

                    if (matched && (flags.find("finish") != flags.end()))
                    {
                        break;
                    }
                }
            }

            if (verbose && new_matches[*pkg_iter].empty())
            {
                fprintf(stderr, "Package \'%s\' not matched\n", pkg_iter->c_str());
            }
        }

        // remove common unchanged records from both maps, remove changed from old map,
        // after that all that remains in old map is removed records,
        // all that remains in new map - updated and new records
        {
            auto iter1 = old_matches.begin();
            auto iter2 = new_matches.begin();

            while ((iter1 != old_matches.end()) && (iter2 != new_matches.end()))
            {
                if (iter1->first < iter2->first)
                {
                    ++iter1;
                }
                else if (iter1->first > iter2->first)
                {
                    ++iter2;
                }
                else
                {
                    if (iter1->second == iter2->second)
                    {
                        iter2 = new_matches.erase(iter2);
                    }
                    else
                    {
                        ++iter2;
                    }

                    iter1 = old_matches.erase(iter1);
                }
            }
        }

        auto pkg_name = std::make_shared<clickhouse::ColumnString>();
        auto cdm_vendor = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto cdm_product = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto cdm_software_platform = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());

        for (auto iter = new_matches.begin(); iter != new_matches.end(); ++iter)
        {
            pkg_name->Append(iter->first);

            auto vendor_array = std::make_shared<clickhouse::ColumnString>();
            auto product_array = std::make_shared<clickhouse::ColumnString>();
            auto software_platform_array = std::make_shared<clickhouse::ColumnString>();

            for (auto match_iter = iter->second.begin(); match_iter != iter->second.end(); ++match_iter)
            {
                vendor_array->Append(std::get<0>(*match_iter));
                product_array->Append(std::get<1>(*match_iter));
                software_platform_array->Append(std::get<2>(*match_iter));
            }

            cdm_vendor->AppendAsColumn(vendor_array);
            cdm_product->AppendAsColumn(product_array);
            cdm_software_platform->AppendAsColumn(software_platform_array);
        }

        if (pkg_name->Size() > 0)
        {
            clickhouse::Block block;

            block.AppendColumn("pkg_name", pkg_name);
            block.AppendColumn("cdm_match.vendor", cdm_vendor);
            block.AppendColumn("cdm_match.product", cdm_product);
            block.AppendColumn("cdm_match.software_platform", cdm_software_platform);

            if (verbose) {
                fprintf(stderr, "Committing data: %zu entries\n", pkg_name->Size());
            }

            client.Insert(table_name_cpe_dictionary_matches, block);
        } else if (verbose) {
            fprintf(stderr, "No new data to commit\n");
        }
    }
    catch (const std::exception &e)
    {
        fprintf(stderr, "Error: %s\n", e.what());
        return EXIT_FAILURE;
    }
    catch (...)
    {
        fprintf(stderr, "Unknown failure\n");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
