/*
 * clickhouse-output.cpp - export data to clickhouse database
 *
 * Copyright (C) 2019 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "clickhouse-output.h"

#include <stdio.h>

#include <algorithm>
#include <cctype>
#include <ctime>
#include <istream>
#include <map>
#include <memory>
#include <set>
#include <sstream>
#include <utility>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>

#include <boost/lexical_cast.hpp>

#include <clickhouse-cpp/client.h>

#include "helpers.h"

#define NVD_CVE_URI "https://nvd.nist.gov/vuln/detail/"

namespace {

const char* issue_status_to_enum_status_string(issue_status status)
{
    const std::map<issue_status, const char*> result_map = {
        { issue_status::actual, "check" },
        { issue_status::patched, "patched" },
        { issue_status::discarded_by_version_check, "discarded_by_version_check" },
        { issue_status::manually_excluded, "manually_excluded" },
        { issue_status::removed, "removed" },
        /* duplicate statuses */
        { issue_status::excluded_as_patched, "patched" },
    };

    auto iter = result_map.find(status);
    if (iter == result_map.end())
    {
        throw std::runtime_error(std::string("Unrecognized issue status value: ") + boost::lexical_cast<std::string>(static_cast<int>(status)));
    }

    return iter->second;
}

const char* actuality_status_to_enum_status_string(actuality_status status)
{
    const std::map<actuality_status, const char*> result_map = {
        { actuality_status::actual, "actual" },
        { actuality_status::removed, "removed" },
    };

    auto iter = result_map.find(status);
    if (iter == result_map.end())
    {
        throw std::runtime_error(std::string("Unrecognized actuality status value: ") + boost::lexical_cast<std::string>(static_cast<int>(status)));
    }

    return iter->second;
}

} // unnamed namespace

void clickhouse_write_report(CveCheckTool &self)
{
    clickhouse::ClientOptions client_options;

    std::string table_name_vulnerabilities_status = "VulnerabilitiesStatus";
    std::string table_name_absent_packages;

    /* First read config */
    {
        boost::property_tree::ptree clickhouse_config_tree;

        std::ifstream stream(self.clickhouse_config);
        boost::property_tree::read_ini(stream, clickhouse_config_tree);

        auto table_names = clickhouse_config_tree.get_child_optional("Tables");
        if (table_names)
        {
            optional_set(table_name_vulnerabilities_status, *table_names, "TableNameVulnerabilitiesStatus", &is_valid_table_name);
            optional_set(table_name_absent_packages, *table_names, "TableNameAbsentPackages", &is_valid_table_name);
        }

        auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
        if (output_connection)
        {
            fill_connection_options(*output_connection, client_options);
        }
        else
        {
            auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
            if (client_connection)
            {
                fill_connection_options(*client_connection, client_options);
            }
        }
    }

    /* if not all packages are scanned, don't update absent packages table */
    if (self.query_type != QueryType::ALL)
    {
        table_name_absent_packages.clear();
    }

    /* Now connect to clickhouse */
    clickhouse::Client client(client_options);

    /* Create tables if they're missing */
    {
        client.Execute(vulnerabilities_status_table_creation_command(table_name_vulnerabilities_status));

        if (not table_name_absent_packages.empty())
        {
            client.Execute(absent_packages_table_creation_command(table_name_absent_packages));
        }
    }

    if (self.verbose) {
        fprintf(stderr, "Clickhouse output function started processing packages\n");
    }

    /* Now insert all data */
    {
        auto pkg_hash = std::make_shared<clickhouse::ColumnUInt64>();
        auto vr_id = std::make_shared<clickhouse::ColumnString>();
        auto vs_status = std::make_shared<clickhouse::ColumnEnum8>(clickhouse::Type::CreateEnum8({{"check", 0}, {"patched", 1}, {"discarded_by_version_check", 2}, {"manually_excluded", 3}, {"removed", 4}}));
        auto vs_cpe = std::make_shared<clickhouse::ColumnString>();
        auto vs_cpe_version_start_excluding = std::make_shared<clickhouse::ColumnString>();
        auto vs_cpe_version_start_including = std::make_shared<clickhouse::ColumnString>();
        auto vs_cpe_version_end_excluding = std::make_shared<clickhouse::ColumnString>();
        auto vs_cpe_version_end_including = std::make_shared<clickhouse::ColumnString>();
        auto cv_id = std::make_shared<clickhouse::ColumnUInt64>();
        auto cv_id_nulls = std::make_shared<clickhouse::ColumnUInt8>();

        auto append_function = [&](uint64_t murmur_hash, const std::string &vulnerability_id, const char *affected_state)
        {
            pkg_hash->Append(murmur_hash);
            vr_id->Append(vulnerability_id);
            vs_status->Append(affected_state);
        };

        auto append_issue_function = [&](const issue_t &issue)
        {
            vs_cpe->Append(issue.cpe.cpe);
            vs_cpe_version_start_excluding->Append(issue.cpe.versionStartExcluding ? *(issue.cpe.versionStartExcluding) : "");
            vs_cpe_version_start_including->Append(issue.cpe.versionStartIncluding ? *(issue.cpe.versionStartIncluding) : "");
            vs_cpe_version_end_excluding->Append(issue.cpe.versionEndExcluding ? *(issue.cpe.versionEndExcluding) : "");
            vs_cpe_version_end_including->Append(issue.cpe.versionEndIncluding ? *(issue.cpe.versionEndIncluding) : "");
            cv_id->Append(issue.exclude_hash ? *(issue.exclude_hash) : 0);
            cv_id_nulls->Append(issue.exclude_hash ? 0 : 1);
        };

        auto append_discarded_issue_function = [&]()
        {
            vs_cpe->Append("");
            vs_cpe_version_start_excluding->Append("");
            vs_cpe_version_start_including->Append("");
            vs_cpe_version_end_excluding->Append("");
            vs_cpe_version_end_including->Append("");
            cv_id->Append(0);
            cv_id_nulls->Append(1);
        };

        auto commit_data = [&]()
        {
            clickhouse::Block block;

            block.AppendColumn("pkg_hash", pkg_hash);
            block.AppendColumn("vr_id", vr_id);
            block.AppendColumn("vs_status", vs_status);
            block.AppendColumn("vs_cpe", vs_cpe);
            block.AppendColumn("vs_cpe_version_start_excluding", vs_cpe_version_start_excluding);
            block.AppendColumn("vs_cpe_version_start_including", vs_cpe_version_start_including);
            block.AppendColumn("vs_cpe_version_end_excluding", vs_cpe_version_end_excluding);
            block.AppendColumn("vs_cpe_version_end_including", vs_cpe_version_end_including);
            block.AppendColumn("cv_id", std::make_shared<clickhouse::ColumnNullable>(cv_id, cv_id_nulls));

            if (self.verbose) {
                fprintf(stderr, "Committing data\n");
            }

            client.Insert(table_name_vulnerabilities_status, block);

            pkg_hash->Clear();
            vr_id->Clear();
            vs_status->Clear();
            vs_cpe->Clear();
            vs_cpe_version_start_excluding->Clear();
            vs_cpe_version_start_including->Clear();
            vs_cpe_version_end_excluding->Clear();
            vs_cpe_version_end_including->Clear();
            cv_id->Clear();
            cv_id_nulls->Clear();
        };

        size_t current = 0;

        for (auto iter = self.db.begin(); iter != self.db.end(); ++iter)
        {
            for (auto issues_iter = iter->second.issues.begin(); issues_iter != iter->second.issues.end(); ++issues_iter)
            {
                ++current;

                if (self.verbose) {
                    fprintf(stderr, "Processing output: %zu, %s-%s-%s, %s, state: check\n", current, iter->first.name.c_str(), iter->first.version.c_str(), iter->first.release.c_str(), issues_iter->vulnerability_id.c_str());
                }

                // remove found cve
                self.existing_vulnerabilities[issues_iter->vulnerability_id][iter->first.murmur_hash].processed = true;

                switch (issues_iter->status) {
                case issue_status::patched:
                case issue_status::excluded_as_patched:
                case issue_status::manually_excluded:
                    throw std::runtime_error(std::string("Unsupported status for actual vulnerability: ") + std::string(issue_status_to_enum_status_string(issues_iter->status)));
                    break;
                }

                append_function(iter->first.murmur_hash, issues_iter->vulnerability_id, "check");
                append_issue_function(*issues_iter);

                if ((self.commit_data_limit) != 0 && (pkg_hash->Size() >= self.commit_data_limit)) {
                    commit_data();
                }
            }

            for (auto patched_issues_iter = iter->second.patched.begin(); patched_issues_iter != iter->second.patched.end(); ++patched_issues_iter)
            {
                ++current;

                if (self.verbose) {
                    fprintf(stderr, "Processing output: %zu, %s-%s-%s, %s, state: patched\n", current, iter->first.name.c_str(), iter->first.version.c_str(), iter->first.release.c_str(), patched_issues_iter->vulnerability_id.c_str());
                }

                // remove found cve
                self.existing_vulnerabilities[patched_issues_iter->vulnerability_id][iter->first.murmur_hash].processed = true;

                switch (patched_issues_iter->status) {
                case issue_status::actual:
                    throw std::runtime_error(std::string("Unsupported status for actual vulnerability: ") + std::string(issue_status_to_enum_status_string(patched_issues_iter->status)));
                    break;
                }

                append_function(iter->first.murmur_hash, patched_issues_iter->vulnerability_id, issue_status_to_enum_status_string(patched_issues_iter->status));
                append_issue_function(*patched_issues_iter);

                if ((self.commit_data_limit) != 0 && (pkg_hash->Size() >= self.commit_data_limit)) {
                    commit_data();
                }
            }

            for (auto discarded_issues_iter = iter->second.discarded_issues.begin(); discarded_issues_iter != iter->second.discarded_issues.end(); ++discarded_issues_iter)
            {
                ++current;

                if (self.verbose) {
                    fprintf(stderr, "Processing output: %zu, %s-%s-%s, %s, state: discarded\n", current, iter->first.name.c_str(), iter->first.version.c_str(), iter->first.release.c_str(), discarded_issues_iter->vulnerability_id.c_str());
                }

                // remove found vulnerability
                self.existing_vulnerabilities[discarded_issues_iter->vulnerability_id][iter->first.murmur_hash].processed = true;

                const char *reason = "check"; // fallback value. must never be used. TODO: consider using special value like 'unknown'.

                switch (discarded_issues_iter->discard_reason)
                {
                case discard_reason_version_mismatch:
                    reason = "discarded_by_version_check";
                    break;

                default:
                    fprintf(stderr, "Unknown reason for discarding %s, using fallback reason %s\n", discarded_issues_iter->vulnerability_id.c_str(), reason);
                    break;
                }

                append_function(iter->first.murmur_hash, discarded_issues_iter->vulnerability_id, reason);
                append_discarded_issue_function();

                if ((self.commit_data_limit) != 0 && (pkg_hash->Size() >= self.commit_data_limit)) {
                    commit_data();
                }
            }
        }

        for (auto existing_vulnerability_iter = self.existing_vulnerabilities.begin(); existing_vulnerability_iter != self.existing_vulnerabilities.end(); ++existing_vulnerability_iter) {
            for (auto hash_iter = existing_vulnerability_iter->second.begin(); hash_iter != existing_vulnerability_iter->second.end(); ++hash_iter) {
                if (hash_iter->second.processed || (hash_iter->second.status == issue_status::removed)) {
                    // entry already processed or marked as removed, skip it
                    continue;
                }

                ++current;

                if (self.verbose) {
                    fprintf(stderr, "Processing output: %zu, %s, state: removed\n", current, existing_vulnerability_iter->first.c_str());
                }

                append_function(hash_iter->first, existing_vulnerability_iter->first, "removed");
                append_discarded_issue_function();

                if ((self.commit_data_limit) != 0 && (pkg_hash->Size() >= self.commit_data_limit)) {
                    commit_data();
                }
            }
        }

        if (pkg_hash->Size() > 0) {
            commit_data();
        }
    }

    if (not table_name_absent_packages.empty())
    {
        if (self.verbose) {
            fprintf(stderr, "Clickhouse output function processing absent packages\n");
        }

        std::set<std::tuple<std::string, std::string, std::string, std::string> > old_absent_products, new_absent_products;

        std::string query_string;

        try
        {
            query_string = std::string(
                        "SELECT ap_vendor_name_ap"
                        ", ap_product_name_ap"
                        ", ap_software_platform_ap"
                        ", vr_id_ap"
                        " FROM ("
                            "SELECT ap_vendor_name AS ap_vendor_name_ap"
                            ", ap_product_name AS ap_product_name_ap"
                            ", ap_software_platform AS ap_software_platform_ap"
                            ", vr_id AS vr_id_ap"
                            ", argMax(ap_status, ts) AS ap_status_ts"
                            " FROM ") + table_name_absent_packages + " GROUP BY ap_vendor_name, ap_product_name, ap_software_platform, vr_id"
                        ") WHERE ap_status_ts = 'actual'";

            /* Now fetch all source packages data */
            client.Select(query_string,
                          [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    auto ap_vendor_name = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                    auto ap_product_name = sv_to_string(ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i));
                    auto ap_software_platform = sv_to_string(ensure_not_null(block[2]->As<clickhouse::ColumnString>())->At(i));
                    auto vr_id = sv_to_string(ensure_not_null(block[3]->As<clickhouse::ColumnString>())->At(i));

                    old_absent_products.insert(std::move(std::make_tuple(std::move(ap_vendor_name), std::move(ap_product_name), std::move(ap_software_platform), std::move(vr_id))));
                }
            });

            query_string.clear();
        }
        catch (...)
        {
            if (!query_string.empty())
            {
                fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
            }

            throw;
        }

        for (auto product_iter = self.cve_db.product_issues_map.begin(); product_iter != self.cve_db.product_issues_map.end(); ++product_iter)
        {
            for (auto vendor_iter = product_iter->second.begin(); vendor_iter != product_iter->second.end(); ++vendor_iter)
            {
                for (auto software_platform_iter = vendor_iter->second.begin(); software_platform_iter != vendor_iter->second.end(); ++software_platform_iter)
                {
                    for (auto vulnerability_iter = software_platform_iter->second.begin(); vulnerability_iter != software_platform_iter->second.end(); ++vulnerability_iter)
                    {
                        new_absent_products.insert(std::make_tuple(vendor_iter->first, product_iter->first, software_platform_iter->first, *vulnerability_iter));
                    }
                }
            }
        }

        {
            auto iter1 = self.product_db.begin();
            auto iter2 = new_absent_products.begin();

            while ((iter1 != self.product_db.end()) && (iter2 != new_absent_products.end()))
            {
                if (*iter1 < std::tie(std::get<0>(*iter2), std::get<1>(*iter2), std::get<2>(*iter2)))
                {
                    ++iter1;
                }
                else if (*iter1 > std::tie(std::get<0>(*iter2), std::get<1>(*iter2), std::get<2>(*iter2)))
                {
                    ++iter2;
                }
                else
                {
                    ++iter1;
                    iter2 = new_absent_products.erase(iter2);
                }
            }
        }

        // remove common records from both sets
        // after that all that remains in old set is removed records,
        // all that remains in new set - new records
        {
            auto iter1 = old_absent_products.begin();
            auto iter2 = new_absent_products.begin();

            while ((iter1 != old_absent_products.end()) && (iter2 != new_absent_products.end()))
            {
                if (*iter1 < *iter2)
                {
                    ++iter1;
                }
                else if (*iter1 > *iter2)
                {
                    ++iter2;
                }
                else
                {
                    iter1 = old_absent_products.erase(iter1);
                    iter2 = new_absent_products.erase(iter2);
                }
            }
        }

        auto ap_vendor_name = std::make_shared<clickhouse::ColumnString>();
        auto ap_product_name = std::make_shared<clickhouse::ColumnString>();
        auto ap_software_platform = std::make_shared<clickhouse::ColumnString>();
        auto vr_id = std::make_shared<clickhouse::ColumnString>();
        auto ap_status = std::make_shared<clickhouse::ColumnEnum8>(clickhouse::Type::CreateEnum8({{"actual", 0}, {"removed", 1}}));

        auto commit_data = [&]()
        {
            clickhouse::Block block;

            block.AppendColumn("ap_vendor_name", ap_vendor_name);
            block.AppendColumn("ap_product_name", ap_product_name);
            block.AppendColumn("ap_software_platform", ap_software_platform);
            block.AppendColumn("vr_id", vr_id);
            block.AppendColumn("ap_status", ap_status);

            if (self.verbose) {
                fprintf(stderr, "Committing data\n");
            }

            client.Insert(table_name_absent_packages, block);

            ap_vendor_name->Clear();
            ap_product_name->Clear();
            ap_software_platform->Clear();
            vr_id->Clear();
            ap_status->Clear();
        };

        for (auto iter = old_absent_products.begin(); iter != old_absent_products.end(); ++iter)
        {
            ap_vendor_name->Append(std::get<0>(*iter));
            ap_product_name->Append(std::get<1>(*iter));
            ap_software_platform->Append(std::get<2>(*iter));
            vr_id->Append(std::get<3>(*iter));
            ap_status->Append("removed");

            if (self.verbose) {
                fprintf(stderr, "Removed absent product: %s %s %s\n", std::get<0>(*iter).c_str(), std::get<1>(*iter).c_str(), std::get<2>(*iter).c_str());
            }

            if ((self.commit_data_limit) != 0 && (ap_vendor_name->Size() >= self.commit_data_limit)) {
                commit_data();
            }
        }

        for (auto iter = new_absent_products.begin(); iter != new_absent_products.end(); ++iter)
        {
            ap_vendor_name->Append(std::get<0>(*iter));
            ap_product_name->Append(std::get<1>(*iter));
            ap_software_platform->Append(std::get<2>(*iter));
            vr_id->Append(std::get<3>(*iter));
            ap_status->Append("actual");

            if (self.verbose) {
                fprintf(stderr, "Added absent product: %s %s %s\n", std::get<0>(*iter).c_str(), std::get<1>(*iter).c_str(), std::get<2>(*iter).c_str());
            }

            if ((self.commit_data_limit) != 0 && (ap_vendor_name->Size() >= self.commit_data_limit)) {
                commit_data();
            }
        }

        if (ap_vendor_name->Size() > 0) {
            commit_data();
        }
    }

    if (self.verbose) {
        fprintf(stderr, "Clickhouse output function finished processing packages\n");
    }

}
