/*
 * main.c - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <algorithm>
#include <memory>
#include <sstream>
#include <string>
#include <utility>

#include <boost/algorithm/string.hpp>
#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>

#include <curl/curl.h>
#include <libgen.h>

#include "cve-check-tool.h"

#include "core.h"

#include "clickhouse-input.h"
#include "clickhouse-output.h"

static bool issue_status_matches(issue_status first, issue_status second)
{
        if (first == issue_status::excluded_as_patched) {
                first = issue_status::patched;
        }

        if (second == issue_status::excluded_as_patched) {
                second = issue_status::patched;
        }

        return (first == second);
}

static void cve_add_package_internal(CveCheckTool &self, source_package_base &pkg_base, const source_package_input_info &input_info)
{
        std::set<issue_t> issues;
        source_package_output_info output_info;

        auto cpe_source_pkg = self.pkgname_to_cpe_vendor_product_softwareplatform.find(pkg_base.name);
        if (cpe_source_pkg != self.pkgname_to_cpe_vendor_product_softwareplatform.end()) {
                for (auto cpe_match_iter = cpe_source_pkg->second.begin(); cpe_match_iter != cpe_source_pkg->second.end(); ++cpe_match_iter) {
                        std::set<issue_t> sourcepkg_issues;

                        sourcepkg_issues = self.cve_db.get_issues(std::get<0>(*cpe_match_iter), std::get<1>(*cpe_match_iter), std::get<2>(*cpe_match_iter), pkg_base.version, self.report_discarded_vulnerabilities ? &(output_info.discarded_issues) : NULL);

                        issues.insert(sourcepkg_issues.begin(), sourcepkg_issues.end());

                        self.product_db.insert(*cpe_match_iter);
                }
        }

        for (auto binary_package_iter = input_info.binary_packages.begin(); binary_package_iter != input_info.binary_packages.end(); ++binary_package_iter)
        {
                auto cpe_binary_pkg = self.pkgname_to_cpe_vendor_product_softwareplatform.find(binary_package_iter->name);
                if (cpe_binary_pkg == self.pkgname_to_cpe_vendor_product_softwareplatform.end()) {
                        continue;
                }

                for (auto cpe_match_iter = cpe_binary_pkg->second.begin(); cpe_match_iter != cpe_binary_pkg->second.end(); ++cpe_match_iter) {
                        std::set<issue_t> binpkg_issues;

                        binpkg_issues = self.cve_db.get_issues(std::get<0>(*cpe_match_iter), std::get<1>(*cpe_match_iter), std::get<2>(*cpe_match_iter), binary_package_iter->version, self.report_discarded_vulnerabilities ? &(output_info.discarded_issues) : NULL);

                        issues.insert(binpkg_issues.begin(), binpkg_issues.end());

                        self.product_db.insert(*cpe_match_iter);
                }
        }

        for (auto issues_iter = issues.begin(); issues_iter != issues.end(); ++issues_iter) {
                issue_status status = issue_status::actual;
                std::optional<uint64_t> exclude_hash;

                auto status_iter = input_info.vulnerabilities_extra_map.find(issues_iter->vulnerability_id);
                if (status_iter != input_info.vulnerabilities_extra_map.end()) {
                        status = status_iter->second.vulnerability_status;
                        exclude_hash = status_iter->second.cv_hash;
                }

                // it looks like one CVE may have multiple statuses due to multiple CPEs. If CVE is both discarded and actual/patched, remove it from discarded ones
                output_info.discarded_issues.erase(discarded_issue_t(issues_iter->vulnerability_id));

                auto cve_iter = self.existing_vulnerabilities.find(issues_iter->vulnerability_id);
                if (cve_iter != self.existing_vulnerabilities.end()) {
                        auto status_iter = cve_iter->second.find(pkg_base.murmur_hash);
                        if (status_iter != cve_iter->second.end()) {
                                status_iter->second.processed = true;
                                if (issue_status_matches(status_iter->second.status, status)) {
                                        // issue already present, skip it
                                        continue;
                                }
                        }
                }

                switch (status) {
                case issue_status::actual:
                        if (std::find(output_info.issues.begin(), output_info.issues.end(), *issues_iter) == output_info.issues.end()) {
                                issue_t issue_copy = *issues_iter;

                                issue_copy.status = status;
                                issue_copy.exclude_hash = exclude_hash;

                                output_info.issues.insert(std::move(issue_copy));
                        }
                        break;

                case issue_status::patched:
                case issue_status::excluded_as_patched:
                case issue_status::manually_excluded:
                        if (std::find(output_info.patched.begin(), output_info.patched.end(), *issues_iter) == output_info.patched.end()) {
                                issue_t issue_copy = *issues_iter;

                                issue_copy.status = status;
                                issue_copy.exclude_hash = exclude_hash;

                                output_info.patched.insert(std::move(issue_copy));
                        }
                        break;
                }
        }

        // check discarded issues
        for (auto discarded_issues_iter = output_info.discarded_issues.begin(); discarded_issues_iter != output_info.discarded_issues.end(); ) {
                auto cve_iter = self.existing_vulnerabilities.find(discarded_issues_iter->vulnerability_id);
                if (cve_iter != self.existing_vulnerabilities.end()) {
                        auto status_iter = cve_iter->second.find(pkg_base.murmur_hash);
                        if (status_iter != cve_iter->second.end()) {
                                status_iter->second.processed = true;
                                if (issue_status_matches(status_iter->second.status, issue_status::discarded_by_version_check)) {
                                        // discarded issue already present, skip it
                                        discarded_issues_iter = output_info.discarded_issues.erase(discarded_issues_iter);
                                        continue;
                                }
                        }
                }

                ++discarded_issues_iter;
        }

        self.db[std::move(pkg_base)] = std::move(output_info);
}

static void show_version(void)
{
        const char *msg =
            "\
" PACKAGE " " PACKAGE_VERSION
            "\n\
Copyright (C) 2015 Intel Corporation\n\
" PACKAGE_NAME
            " is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.";
        fprintf(stderr, "%s\n", msg);
}

int main(int argc, char **argv)
{
        try
        {
                bool show_help = false;
                bool _show_version = false;
                bool verbose = false;
                uint64_t commit_data_limit = 0;
                std::string clickhouse_config_file;
                bool report_discarded_vulnerabilities = false;
                std::string vendors_filename;
                std::string src_pkg_name;
                std::string src_pkg_version;
                std::string task_id;

                boost::program_options::options_description desc("General Options");
                desc.add_options()
                    ("help,h", boost::program_options::bool_switch(&show_help)->default_value(false), "Show help")
                    ("version,v", boost::program_options::bool_switch(&_show_version)->default_value(false), "Show version")
                    ("verbose,V", boost::program_options::bool_switch(&verbose)->default_value(false), "Very verbose mode")
                    ("report-discarded-vulnerabilities,R", boost::program_options::bool_switch(&report_discarded_vulnerabilities)->default_value(false), "Provide information about discarded vulnerabilities")
                    ("commit-data-limit,L", boost::program_options::value< uint64_t >(&commit_data_limit), "Commit vulnerabilities data when it reaches specified amount of records")
                    ("clickhouse-config,Y", boost::program_options::value< std::string >(&clickhouse_config_file), "Path to file containing configuration for connection to clickhouse database")
                    ("src-pkg-name,S", boost::program_options::value< std::string >(&src_pkg_name), "Name of source package")
                    ("src-pkg-version,s", boost::program_options::value< std::string >(&src_pkg_version), "EVR of source package")
                    ("task-id,T", boost::program_options::value< std::string >(&task_id), "ID of task to be processed. Takes priority over name of source package if specified.")
                    ;

                boost::program_options::options_description hidden;
                hidden.add_options()
                    ("vendors-filename", boost::program_options::value< std::string >(&vendors_filename), "Path to file containing list of explicit vendors");

                boost::program_options::options_description all_options;
                all_options.add(desc);
                all_options.add(hidden);

                boost::program_options::positional_options_description positional_options;
                positional_options.add("vendors-filename", 1);

                boost::program_options::variables_map vm;

                boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(positional_options).run(), vm);
                boost::program_options::notify(vm);

                if (show_help) {
                        std::stringstream descr;
                        descr << all_options;
                        printf("Usage: %s [options]\n%s\n", argv[0], descr.str().c_str());
                        return EXIT_SUCCESS;
                }

                if (_show_version) {
                        show_version();
                        return EXIT_SUCCESS;
                }

                if (clickhouse_config_file.empty()) {
                        fprintf(stderr, "Clickhouse config is required\n");
                        return EXIT_FAILURE;
                }

                CveCheckTool instance;

                instance.commit_data_limit = commit_data_limit;
                instance.clickhouse_config = clickhouse_config_file;
                instance.verbose = verbose;
                instance.report_discarded_vulnerabilities = report_discarded_vulnerabilities;

                instance.src_pkg_name = src_pkg_name;
                instance.src_pkg_version = src_pkg_version;
                instance.task_id = task_id;

                if (!task_id.empty())
                {
                        instance.query_type = QueryType::TASK_ID;
                }
                else if (!src_pkg_name.empty())
                {
                        instance.query_type = QueryType::SRC_RPM;
                }
                else
                {
                        instance.query_type = QueryType::ALL;
                }

                instance.cve_db.load(clickhouse_config_file, verbose);

                scan_package_vendor_product_matching(instance);

                if (!vendors_filename.empty())
                {
                        // read explicit vendors list
                        std::ifstream list_stream(vendors_filename);
                        std::string vendor_line;

                        while (std::getline(list_stream, vendor_line))
                        {
                            if (vendor_line.empty() || (vendor_line[0] == '#'))
                            {
                                continue;
                            }

                            instance.cve_db.explicit_vendors.insert(vendor_line);
                        }
                }

                auto packages = package_scan_all_packages(instance);

                scan_existing_packages_and_vulnerabilities(instance, packages);

                size_t total = packages.size();
                size_t current = 0;

                for (auto iter = packages.begin(); iter != packages.end(); ++iter)
                {
                        ++current;

                        std::string name;

                        if (instance.verbose) {
                                name = std::get<0>(*iter).name;
                        }

                        cve_add_package_internal(instance, std::get<0>(*iter), std::get<1>(*iter));

                        if (instance.verbose) {
                                fprintf(stderr, "Processed input package %zu/%zu: %s\n", current, total, name.c_str());
                        }
                }

                size_t size = instance.db.size();
                if (size == 0) {
                        fprintf(stderr, "No source files were encountered, aborting\n");
                        return EXIT_FAILURE;
                }

                if (instance.verbose) {
                        fprintf(stderr, "Scanned %zu source file%s\n", size, size > 1 ? "s" : "");
                }

                clickhouse_write_report(instance);
        }
        catch (const std::exception &e)
        {
                fprintf(stderr, "Error: %s\n", e.what());
                return EXIT_FAILURE;
        }
        catch (...)
        {
                fprintf(stderr, "Unknown failure\n");
                return EXIT_FAILURE;
        }

        return EXIT_SUCCESS;
}
