/*
 * structs.h - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 * Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#pragma once

#include <stdint.h>

#include <list>
#include <map>
#include <set>

#include "common.h"

struct cve_entry_t {
        std::string id;           /**<CVE ID */

        std::map<std::string, std::map<std::string, std::map<std::string, std::set<cpe_info> > > > cpe;

        bool operator==(const cve_entry_t &other) const;
        bool operator<(const cve_entry_t &other) const;
};

struct binary_package_t {
        std::string name;
        std::string version;
        std::string release;

        binary_package_t() = default;
        binary_package_t(const std::string &l_name, const std::string &l_version, const std::string &l_release);

        bool operator<(const binary_package_t &other) const;
};

enum class issue_status {
        actual = 0,
        patched,
        discarded_by_version_check,
        manually_excluded,
        removed,
        /* duplicate statuses */
        excluded_as_patched,
};

struct issue_t {
        std::string vulnerability_id;
        std::optional<uint64_t> exclude_hash;
        issue_status status = issue_status::actual;

        cpe_info cpe;

        bool operator==(const issue_t &other) const;
        bool operator<(const issue_t &other) const;
};

typedef enum discard_reason {
        discard_reason_version_mismatch = 0,
} discard_reason_t;

struct discarded_issue_t {
        std::string vulnerability_id;
        discard_reason_t discard_reason = discard_reason_version_mismatch;

        discarded_issue_t() = default;
        explicit discarded_issue_t(const std::string &id)
                : vulnerability_id(id)
        {
        }

        bool operator==(const discarded_issue_t &other) const;
        bool operator<(const discarded_issue_t &other) const;
};

struct vulnerability_extra
{
    issue_status vulnerability_status = issue_status::actual;
    std::optional<uint64_t> cv_hash;

    vulnerability_extra() = default;

    explicit vulnerability_extra(issue_status status, const std::optional<uint64_t> &l_cv_hash = std::nullopt)
        : vulnerability_status(status)
        , cv_hash(l_cv_hash)
    {
    }
};

/**
 * Representation of a source package
 */
struct source_package_base {
        uint64_t murmur_hash = 0;
        std::string name;       /**<Our name for the package */
        std::string version;    /**<Our current version */
        std::string release;      /**<Highest (available) release number */

        bool operator==(const source_package_base &other) const;
        bool operator<(const source_package_base &other) const;
};

struct source_package_input_info {
        std::map<std::string, vulnerability_extra> vulnerabilities_extra_map;
        std::set<binary_package_t> binary_packages; /**<Optional: list of binary packages built from this source package */
};

struct source_package_output_info {
        std::set<issue_t> issues;    /**<List of applicable (non-patched) CVEs */
        std::set<issue_t> patched;   /**<List of *patched* applicable CVEs */
        std::set<discarded_issue_t> discarded_issues; /**<Optional list of discarded_issue_t */
};
