/*
 * update.c - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <algorithm>
#include <cmath>
#include <limits>
#include <string>
#include <utility>

#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>

#include <clickhouse-cpp/client.h>

#include "cve-db-lock.h"
#include "helpers.h"
#include "parse-json.h"
#include "util.h"
#include "update.h"

static void show_version(void)
{
        const char *msg =
            "\
" PACKAGE " " PACKAGE_VERSION
            "\n\
Copyright (C) 2015 Intel Corporation\n\
" PACKAGE_NAME
            " is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.";
        fprintf(stderr, "%s\n", msg);
}

int main(int argc, char **argv)
{
    try
    {
        bool show_help = false;
        bool _show_version = false;
        bool verbose = false;
        bool skip_download = false;
        std::string nvds;
        std::string _cacert_file;
        std::string clickhouse_config_file;
        uint64_t commit_data_limit = 0;

        boost::program_options::options_description desc("General Options");
        desc.add_options()
                ("help,h", boost::program_options::bool_switch(&show_help)->default_value(false), "Show help")
                ("version,v", boost::program_options::bool_switch(&_show_version)->default_value(false), "Show version")
                ("verbose,V", boost::program_options::bool_switch(&verbose)->default_value(false), "Very verbose mode")
                ("skip-download,U", boost::program_options::bool_switch(&skip_download)->default_value(false), "Bypass forced file downloads")
                ("nvd-dir,d", boost::program_options::value< std::string >(&nvds), "NVD directory in filesystem")
                ("cacert,C", boost::program_options::value< std::string >(&_cacert_file), "Path to the combined SSL certificates file (system default is used if not set)")
                ("commit-data-limit,L", boost::program_options::value< uint64_t >(&commit_data_limit), "Commit vulnerabilities data when it reaches specified amount of records")
                ("clickhouse-config,Y", boost::program_options::value< std::string >(&clickhouse_config_file), "Path to file containing configuration for connection to clickhouse database");

        boost::program_options::options_description all_options;
        all_options.add(desc);

        boost::program_options::variables_map vm;

        boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).run(), vm);
        boost::program_options::notify(vm);

        if (show_help) {
            std::stringstream descr;
            descr << all_options;
            printf("Usage: %s [options]\n%s\n", argv[0], descr.str().c_str());
            return EXIT_SUCCESS;
        }

        if (_show_version) {
            show_version();
            return EXIT_SUCCESS;
        }

        if (clickhouse_config_file.empty()) {
            fprintf(stderr, "Clickhouse config is required\n");
            return EXIT_FAILURE;
        }

        std::map<std::string, vulnerability_record> current_vulnerabilities, new_vulnerabilities;

        std::string db_path = get_db_path(nvds);

        CveDbLock db_lock(db_path);

        bool db_locked = db_lock.cve_db_read_lock(CveDbLock::LOCK_WAIT_SECS);
        if (!db_locked) {
            fputs("Exiting...\n", stderr);
            return EXIT_FAILURE;
        }

        bool update_needed = update_required(db_path);

        if (update_needed && !skip_download) {
            fprintf(stderr, "Update of db forced\n");
        }

        db_lock.cve_db_unlock();
        if (!update_db(db_lock, verbose, skip_download || !update_needed, db_path, _cacert_file, [&new_vulnerabilities](const std::string &filename) {
                load_vulnerabilities_json(filename, new_vulnerabilities);
            })) {
            fprintf(stderr, "Failed to update database\n");
            return EXIT_FAILURE;
        }

        clickhouse::ClientOptions client_options;

        std::string table_name_vulnerabilities = "Vulnerabilities";

        /* First read config */
        {
            boost::property_tree::ptree clickhouse_config_tree;

            std::ifstream stream(clickhouse_config_file);
            boost::property_tree::read_ini(stream, clickhouse_config_tree);

            auto table_names = clickhouse_config_tree.get_child_optional("Tables");
            if (table_names)
            {
                optional_set(table_name_vulnerabilities, *table_names, "TableNameVulnerabilities", &is_valid_table_name);
            }

            auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
            if (output_connection)
            {
                fill_connection_options(*output_connection, client_options);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options);
                }
            }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        std::string query_string;

        /* Create tables if they're missing */
        client.Execute(vulnerabilities_table_creation_command(table_name_vulnerabilities));

        if (verbose) {
            fprintf(stderr, "Obtaining information about existing vulnerabilities\n");
        }

        try
        {
            query_string = std::string(
                        "SELECT * FROM ("
                            "SELECT vr_id"
                            ", argMax(vr_description, ts)"
                            ", argMax(vr_url, ts)"
                            ", argMax(vr_score, ts)"
                            ", argMax(vr_attacktype, ts)"
                            ", argMax(vr_uris, ts)"
                            ", argMax(vr_modifieddate, ts)"
                            ", argMax(vr_cpe.string, ts)"
                            ", argMax(vr_cpe.version_start_excluding, ts)"
                            ", argMax(vr_cpe.version_start_including, ts)"
                            ", argMax(vr_cpe.version_end_excluding, ts)"
                            ", argMax(vr_cpe.version_end_including, ts)"
                            ", argMax(vr_configurations, ts)"
                            ", argMax(vr_type, ts)"
                            ", argMax(vr_status, ts) AS vr_status_ts"
                            " FROM ") + table_name_vulnerabilities + " GROUP BY vr_id"
                        ") WHERE vr_status_ts = 'actual'";

            /* Now fetch all source packages data */
            client.Select(query_string,
                          [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    vulnerability_record vulnerability;

                    auto vr_id = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                    vulnerability.description = ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i);
                    vulnerability.url = ensure_not_null(block[2]->As<clickhouse::ColumnString>())->At(i);
                    vulnerability.score = ensure_not_null(block[3]->As<clickhouse::ColumnFloat64>())->At(i);
                    vulnerability.attacktype = ensure_not_null(block[4]->As<clickhouse::ColumnString>())->At(i);

                    auto vr_uris_entries = ensure_not_null(ensure_not_null(block[5]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());

                    for (size_t j = 0; j < vr_uris_entries->Size(); ++j)
                    {
                        vulnerability.uris.insert(sv_to_string(vr_uris_entries->At(j)));
                    }

                    vulnerability.modifieddate = ensure_not_null(block[6]->As<clickhouse::ColumnDateTime>())->At(i);

                    auto vr_cpe_strings = ensure_not_null(ensure_not_null(block[7]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto vr_cpe_start_excluding = ensure_not_null(ensure_not_null(block[8]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto vr_cpe_start_including = ensure_not_null(ensure_not_null(block[9]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto vr_cpe_end_excluding = ensure_not_null(ensure_not_null(block[10]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto vr_cpe_end_including = ensure_not_null(ensure_not_null(block[11]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());

                    if ((vr_cpe_strings->Size() != vr_cpe_start_excluding->Size())
                            || (vr_cpe_strings->Size() != vr_cpe_start_including->Size())
                            || (vr_cpe_strings->Size() != vr_cpe_end_excluding->Size())
                            || (vr_cpe_strings->Size() != vr_cpe_end_including->Size()))
                    {
                        std::stringstream str;
                        str << "Columns vr_cpe.string, vr_cpe.version_start_excluding, vr_cpe.version_start_including, vr_cpe.version_end_excluding, vr_cpe.version_end_including have mismatching sizes!";
                        throw std::runtime_error(str.str());
                    }

                    for (size_t j = 0; j < vr_cpe_strings->Size(); ++j)
                    {
                        cpe_info info;
                        info.cpe = vr_cpe_strings->At(j);

                        auto tmp_value = vr_cpe_start_excluding->At(j);
                        if (!tmp_value.empty())
                        {
                            info.versionStartExcluding = tmp_value;
                        }

                        tmp_value = vr_cpe_start_including->At(j);
                        if (!tmp_value.empty())
                        {
                            info.versionStartIncluding = tmp_value;
                        }

                        tmp_value = vr_cpe_end_excluding->At(j);
                        if (!tmp_value.empty())
                        {
                            info.versionEndExcluding = tmp_value;
                        }

                        tmp_value = vr_cpe_end_including->At(j);
                        if (!tmp_value.empty())
                        {
                            info.versionEndIncluding = tmp_value;
                        }

                        vulnerability.cpe.insert(std::move(info));
                    }

                    vulnerability.configurations = ensure_not_null(block[12]->As<clickhouse::ColumnString>())->At(i);
                    vulnerability.type = ensure_not_null(block[13]->As<clickhouse::ColumnString>())->At(i);
                    vulnerability.status = static_cast<actuality_status>(ensure_not_null(block[14]->As<clickhouse::ColumnEnum8>())->At(i));

                    current_vulnerabilities[vr_id] = std::move(vulnerability);
                }
            });

            query_string.clear();
        }
        catch (...)
        {
            if (!query_string.empty())
            {
                fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
            }

            throw;
        }

        // remove common unchanged vulnerabilities from both maps, remove changed from old map,
        // after that all that remains in old map is removed vulnerabilities,
        // all that remains in new map - updated and new vulnerabilities
        {
            auto iter1 = current_vulnerabilities.begin();
            auto iter2 = new_vulnerabilities.begin();

            while ((iter1 != current_vulnerabilities.end()) && (iter2 != new_vulnerabilities.end()))
            {
                if (iter1->first < iter2->first)
                {
                    ++iter1;
                }
                else if (iter1->first > iter2->first)
                {
                    ++iter2;
                }
                else
                {
                    if (iter1->second == iter2->second)
                    {
                        iter2 = new_vulnerabilities.erase(iter2);
                    }
                    else
                    {
                        ++iter2;
                    }

                    iter1 = current_vulnerabilities.erase(iter1);
                }
            }
        }

        if (verbose)
        {
            fprintf(stderr, "Removed vulnerabilities: %zu, new/updated vulnerabilities: %zu\n", current_vulnerabilities.size(), new_vulnerabilities.size());
        }

        auto vr_id = std::make_shared<clickhouse::ColumnString>();
        auto vr_description = std::make_shared<clickhouse::ColumnString>();
        auto vr_url = std::make_shared<clickhouse::ColumnString>();
        auto vr_score = std::make_shared<clickhouse::ColumnFloat64>();
        auto vr_attacktype = std::make_shared<clickhouse::ColumnString>();
        auto vr_uris = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto vr_modifieddate = std::make_shared<clickhouse::ColumnDateTime>();
        auto vr_cpe_string = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto vr_cpe_version_start_excluding = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto vr_cpe_version_start_including = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto vr_cpe_version_end_excluding = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto vr_cpe_version_end_including = std::make_shared<clickhouse::ColumnArray>(std::make_shared<clickhouse::ColumnString>());
        auto vr_configurations = std::make_shared<clickhouse::ColumnString>();
        auto vr_type = std::make_shared<clickhouse::ColumnString>();
        auto vr_status = std::make_shared<clickhouse::ColumnEnum8>(clickhouse::Type::CreateEnum8({{"actual", 0}, {"removed", 1}}));

        auto append_function = [&](const std::string &vulnerability_id, const vulnerability_record &record, const char *state)
        {
            vr_id->Append(vulnerability_id);
            vr_description->Append(record.description);
            vr_url->Append(record.url);
            vr_score->Append(record.score);
            vr_attacktype->Append(record.attacktype);

            auto vulnerability_uris_value = std::make_shared<clickhouse::ColumnString>();

            for (auto uri_iter = record.uris.begin(); uri_iter != record.uris.end(); ++uri_iter)
            {
                vulnerability_uris_value->Append(*uri_iter);
            }

            vr_uris->AppendAsColumn(vulnerability_uris_value);
            vr_modifieddate->Append(record.modifieddate);

            auto vulnerability_cpe_string = std::make_shared<clickhouse::ColumnString>();
            auto vulnerability_version_start_excluding = std::make_shared<clickhouse::ColumnString>();
            auto vulnerability_version_start_including = std::make_shared<clickhouse::ColumnString>();
            auto vulnerability_version_end_excluding = std::make_shared<clickhouse::ColumnString>();
            auto vulnerability_version_end_including = std::make_shared<clickhouse::ColumnString>();

            for (auto cpe_iter = record.cpe.begin(); cpe_iter != record.cpe.end(); ++cpe_iter)
            {
                vulnerability_cpe_string->Append(cpe_iter->cpe);
                vulnerability_version_start_excluding->Append(cpe_iter->versionStartExcluding ? *(cpe_iter->versionStartExcluding) : std::string());
                vulnerability_version_start_including->Append(cpe_iter->versionStartIncluding ? *(cpe_iter->versionStartIncluding) : std::string());
                vulnerability_version_end_excluding->Append(cpe_iter->versionEndExcluding ? *(cpe_iter->versionEndExcluding) : std::string());
                vulnerability_version_end_including->Append(cpe_iter->versionEndIncluding ? *(cpe_iter->versionEndIncluding) : std::string());
            }

            vr_cpe_string->AppendAsColumn(vulnerability_cpe_string);
            vr_cpe_version_start_excluding->AppendAsColumn(vulnerability_version_start_excluding);
            vr_cpe_version_start_including->AppendAsColumn(vulnerability_version_start_including);
            vr_cpe_version_end_excluding->AppendAsColumn(vulnerability_version_end_excluding);
            vr_cpe_version_end_including->AppendAsColumn(vulnerability_version_end_including);

            vr_configurations->Append(record.configurations);
            vr_type->Append(record.type);
            vr_status->Append(state);
        };

        auto commit_data = [&]()
        {
            clickhouse::Block block;

            block.AppendColumn("vr_id", vr_id);
            block.AppendColumn("vr_description", vr_description);
            block.AppendColumn("vr_url", vr_url);
            block.AppendColumn("vr_score", vr_score);
            block.AppendColumn("vr_attacktype", vr_attacktype);
            block.AppendColumn("vr_uris", vr_uris);
            block.AppendColumn("vr_modifieddate", vr_modifieddate);
            block.AppendColumn("vr_cpe.string", vr_cpe_string);
            block.AppendColumn("vr_cpe.version_start_excluding", vr_cpe_version_start_excluding);
            block.AppendColumn("vr_cpe.version_start_including", vr_cpe_version_start_including);
            block.AppendColumn("vr_cpe.version_end_excluding", vr_cpe_version_end_excluding);
            block.AppendColumn("vr_cpe.version_end_including", vr_cpe_version_end_including);
            block.AppendColumn("vr_configurations", vr_configurations);
            block.AppendColumn("vr_type", vr_type);
            block.AppendColumn("vr_status", vr_status);

            if (verbose) {
                fprintf(stderr, "Committing data\n");
            }

            client.Insert(table_name_vulnerabilities, block);

            vr_id->Clear();
            vr_description->Clear();
            vr_url->Clear();
            vr_score->Clear();
            vr_attacktype->Clear();
            vr_uris->Clear();
            vr_modifieddate->Clear();
            vr_cpe_string->Clear();
            vr_cpe_version_start_excluding->Clear();
            vr_cpe_version_start_including->Clear();
            vr_cpe_version_end_excluding->Clear();
            vr_cpe_version_end_including->Clear();
            vr_configurations->Clear();
            vr_type->Clear();
            vr_status->Clear();
        };

        for (auto iter = current_vulnerabilities.begin(); iter != current_vulnerabilities.end(); ++iter)
        {
            append_function(iter->first, iter->second, "removed");

            if ((commit_data_limit) != 0 && (vr_id->Size() >= commit_data_limit)) {
                commit_data();
            }
        }

        for (auto iter = new_vulnerabilities.begin(); iter != new_vulnerabilities.end(); ++iter)
        {
            append_function(iter->first, iter->second, "actual");

            if ((commit_data_limit) != 0 && (vr_id->Size() >= commit_data_limit)) {
                commit_data();
            }
        }

        if (vr_id->Size() > 0) {
            commit_data();
        }
    }
    catch (const std::exception &e)
    {
        fprintf(stderr, "Error: %s\n", e.what());
        return EXIT_FAILURE;
    }
    catch (...)
    {
        fprintf(stderr, "Unknown failure\n");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
