/*
 * cve-check-tool.c
 *
 * Copyright (C) 2015-2016 Sergey Popovich <popovich_sergei@mail.ua>.
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "cve-db-lock.h"
#include "util.h"

#include <map>
#include <sstream>
#include <stdexcept>

static const std::map<CveDbLock::locktype, short int> locktype2l_type = {
    { CveDbLock::LT_READ, F_RDLCK },
    { CveDbLock::LT_WRITE, F_WRLCK },
};

static const std::map<CveDbLock::locktype, const char*> locktype2string = {
    { CveDbLock::LT_READ, "read" },
    { CveDbLock::LT_WRITE, "write" },
};

#ifndef O_NOFOLLOW
#define O_NOFOLLOW 0
#endif

CveDbLock::CveDbLock(const std::string &db_path)
{
        if (db_path.empty())
        {
                throw std::runtime_error("CveDbLock: db_path is empty");
        }

        db_lock_fname = make_db_dot_fname(db_path, "cve.lock");

        const int flags = O_RDWR | O_CREAT | O_NONBLOCK | O_NOFOLLOW;
        const mode_t mode = S_IRUSR | S_IWUSR;

        db_lock_fd = open(db_lock_fname.c_str(), flags, mode);
        if (db_lock_fd < 0) {
                std::stringstream str;
                str << "Failed to obtain lockfile " << db_lock_fname;
                throw std::runtime_error(str.str());
        }
}

CveDbLock::~CveDbLock()
{
        close(db_lock_fd);
        unlink(db_lock_fname.c_str());
}

bool CveDbLock::cve_db_read_lock(int wait)
{
        return cve_db_lock(LT_READ, wait);
}

bool CveDbLock::cve_db_write_lock(int wait)
{
        return cve_db_lock(LT_WRITE, wait);
}

bool CveDbLock::cve_db_lock(locktype lt, int wait)
{
        const char *lt_str;
        unsigned int waited;

        lt_str = locktype2string.at(lt);

        if (wait < 0) {
                waited = wait = 2;
        } else {
                waited = 0;
        }

        for (;;) {
                struct flock fl = {
                        .l_type = locktype2l_type.at(lt), .l_whence = SEEK_SET,
                };
                int ret;

                ret = fcntl(db_lock_fd, F_SETLK, &fl);
                if (!ret) {
                        return true;
                }
                if (errno != EAGAIN && errno != EACCES) {
                        fprintf(stderr, "Error acquiring database lock: %s\n", strerror(errno));
                        break;
                }

                if (waited % 2 == 0) {
                        fputs("Another app holds the lock on database", stderr);
                        if (wait) {
                                int remaining = wait - waited;
                                if (remaining <= 0) {
                                        fprintf(stderr, "; %s lock is not acquired\n", lt_str);
                                        break;
                                }
                                fprintf(stderr, "; acquiring %s lock within %ds ...", lt_str, remaining);
                        } else {
                                fputs("; waiting indefinitely", stderr);
                        }
                        fputc('\n', stderr);
                }

                sleep(1);
                waited++;
                if (wait && waited >= (unsigned int)wait) {
                        /* last round: make it even */
                        waited = (wait + 1) & ~1;
                }
        }

        return false;
}

void CveDbLock::cve_db_unlock()
{
        struct flock fl = {
                .l_type = F_UNLCK, .l_whence = SEEK_SET,
        };
        int ret;

        ret = fcntl(db_lock_fd, F_SETLK, &fl);

        if (ret != 0)
        {
                std::stringstream str;
                str << "Failed to unlock lockfile " << db_lock_fname;
                throw std::runtime_error(str.str());
        }
}
