/*
 * util.c - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "util.h"

#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <filesystem>
#include <iomanip>
#include <memory>
#include <sstream>

#include <boost/algorithm/string.hpp>

time_t parse_date(const std::string &date)
{
        std::tm t = {};

        /* Example date strings:
         * 2017-08-08T01:31Z
         * 2021-10-19T19:34Z
         * 2021-10-20T14:14Z
         * 2021-07-07T14:28Z
         */

        auto result = strptime(date.c_str(), "%Y-%m-%dT%H:%MZ", &t);
        if ((result == nullptr) || (*result != '\0'))
        {
            std::stringstream str;
            str << "Error while parsing date/time: " << date;
            throw std::runtime_error(str.str());
        }

        time_t time_result = mktime(&t);
        if (time_result == (time_t) -1) {
                std::stringstream str;
                str << "Error while parsing date/time: " << date;
                throw std::runtime_error(str.str());
        }

        return time_result + t.tm_gmtoff;
}

/* File functions */

bool cve_file_exists(const std::string &p)
{
        struct stat st = {.st_ino = 0 };
        return (stat(p.c_str(), &st) == 0);
}

bool cve_is_dir(const std::string &p)
{
        struct stat st = {.st_ino = 0 };
        if (stat(p.c_str(), &st) != 0) {
                return false;
        }
        return S_ISDIR(st.st_mode);
}

std::string cve_get_file_parent(const std::string &p)
{
        std::unique_ptr<char[], void(*)(void*)> real(realpath(p.c_str(), nullptr), &std::free);
        if (!real) {
                std::stringstream str;
                str << "Failed to get file parent for " << p;
                throw std::runtime_error(str.str());
        }

        return std::filesystem::path(real.get()).parent_path().native();
}

std::string make_db_dot_fname(const std::string &db_path, const std::string &suffix)
{
        std::string dir;

        dir.reserve(db_path.length() + sizeof("/.nvd.db.") - 1 + suffix.length());
        dir.append(db_path);
        dir.append("/.nvd.db.");
        dir.append(suffix);

        return dir;
}

std::string nvdcve_make_fname(int year, const std::string &fext)
{
        std::stringstream str;

        switch (year) {
        case -2:
                str << "nvdcve-1.1-recent." << fext;
                break;
        case -1:
                str << "nvdcve-1.1-modified." << fext;
                break;
        default:
                str << "nvdcve-1.1-" << year << "." << fext;
        }

        return str.str();
}

void split_cpe(const std::string &cpe, std::vector<std::string> &result)
{
        std::string::size_type begin = 0, end = 0;

        while ((end = cpe.find(':', end)) != std::string::npos) {
                if ((end == 0) || (cpe[end - 1] != '\\')) {
                        result.push_back(boost::replace_all_copy(cpe.substr(begin, end - begin), "\\", ""));
                        begin = end + 1;
                }

                ++end;
        }

        result.push_back(boost::replace_all_copy(cpe.substr(begin), "\\", ""));
}
